package ru.bitel.bgbilling.modules.inet.dyn.device.redback;

import org.apache.log4j.Logger;

import ru.bitel.bgbilling.kernel.event.EventProcessor;
import ru.bitel.bgbilling.kernel.network.radius.RadiusAttributeSet;
import ru.bitel.bgbilling.kernel.network.radius.RadiusPacket;
import ru.bitel.bgbilling.modules.inet.access.InetConnectionManager;
import ru.bitel.bgbilling.modules.inet.access.sa.ServiceActivator;
import ru.bitel.bgbilling.modules.inet.access.sa.ServiceActivatorEvent;
import ru.bitel.bgbilling.modules.inet.api.common.bean.InetConnection;
import ru.bitel.bgbilling.modules.inet.api.common.bean.InetDevice;
import ru.bitel.bgbilling.modules.inet.api.common.bean.InetDeviceType;
import ru.bitel.bgbilling.modules.inet.api.common.bean.InetServ;
import ru.bitel.bgbilling.modules.inet.dyn.device.radius.AbstractRadiusServiceActivator;
import ru.bitel.bgbilling.server.util.Setup;
import ru.bitel.common.ParameterMap;

public class SmartEdgeServiceActivator
    extends AbstractRadiusServiceActivator
    implements ServiceActivator
{
	private static final Logger logger = Logger.getLogger( SmartEdgeServiceActivator.class );
	
	/**
	 * Ничего не делать для закрытия соединения.
	 * @see #disableServicesOnClose
	 */
	protected static final int CLOSE_MODE_NONE = 1;

	/**
	 * Посылать PoD пакет для закрытия соединения.
	 * @see #disableServicesOnClose
	 */
	protected static final int CLOSE_MODE_POD = 2;
	
	/**
	 * Режим закрытия соединения.
	 * @see #CLOSE_MODE_NONE
	 * @see #CLOSE_MODE_POD
	 */
	protected int closeMode;
	
	/**
	 * Режим закрытия соединения для переключения из отключен во включен (если {@link #withoutBreak}=false).
	 * @see #CLOSE_MODE_NONE
	 * @see #CLOSE_MODE_POD
	 * @see #CLOSE_MODE_SUBSCR_COMMAND
	 */
	protected int closeEnableMode;
	
	/**
	 * Режим закрытия соединения для переключения из включен в отключен (если {@link #withoutBreak}=false).
	 * @see #CLOSE_MODE_NONE
	 * @see #CLOSE_MODE_POD
	 * @see #CLOSE_MODE_SUBSCR_COMMAND
	 */
	protected int closeDisableMode;
	
	/**
	 * Нужно ли посылать CoA при переводе из disable в enable (при withoutBreak=false)
	 */
	protected boolean coaOnEnable;
	
	/**
	 * Нужно ли закрывать сервисы при закрытии сессии.
	 */
	protected boolean disableServicesOnClose;

	/**
	 * Нужно ли удалять из keymap вторичной авторизации. keymap используется в InetDhcpHelperProcessor и InetRadiusHelperProcessor.
	 * При удалении из keymap InetDhcpHelperProcessor начнет выдавать NAK, InetRadiusHelperProcessor перестанет авторизовать.
	 * Для работы с ISG+DHCP должен быть true.
	 */
	protected boolean closeRemoveFromKeyMap;

	/**
	 * Набор атрибутов для закрытия всех сервисов sa.radius.service.closeAttributes=
	 */
	protected RadiusAttributeSet serviceCloseAttributes;
	
	/**
	 * @see #closeMode
	 */
	private final int defaultCloseMode;
	/**
	 * @see #closeEnableMode
	 */
	private final int defaultCloseEnableMode;
	/**
	 * @see #coaOnEnable
	 */
	private final boolean defaultCoaOnEnable;
	/**
	 * @see #closeRemoveFromKeyMap
	 */
	private final boolean defaultCloseRemoveFromKeyMap;
	/**
	 * @see #disableServicesOnClose
	 */
	private final boolean defaultDisableServicesOnClose;
	
	public SmartEdgeServiceActivator()
	{
		// параметры по умолчанию
		this(
			   false, // defaultWithoutBreak - сбрасываем соединение при переключении состояния
			   CLOSE_MODE_POD, // defaultCloseMode - посылаем PoD при необходимости сбросить соединение
			   CLOSE_MODE_POD, // defaultCloseEnableMode - посылаем PoD пакет при необходимости сбросить соединение при переключении из состояния отключен во включен (при withoutBreak=false)
			   false, // defaultCloseEnableMode - по умолчанию не включаем сервисы при необходимости переключить из состояния отключен во включен (при withoutBreak=false)
			   false, // defaultDisableServicesOnClose - по умолчанию не отключаем сервисы когда нужно сбросить соединение
			   false // defaultCloseRemoveFromKeyMap - по умолчанию не выдаем NAK, когда нужно сбросить соединение
		);
	}
	
	protected SmartEdgeServiceActivator( boolean defaultWithoutBreak )
	{
		this( defaultWithoutBreak, CLOSE_MODE_POD, CLOSE_MODE_POD, false, false, false );
	}

	protected SmartEdgeServiceActivator( boolean defaultWithoutBreak, int defaultCloseMode, int defaultCloseEnableMode,
										 boolean defaultCoaOnEnable, boolean defaultDisableServicesOnClose, boolean defaultCloseRemoveFromKeyMap )
	{
		super( "sa.radius.option.", defaultWithoutBreak, "Acct-Session-Id", false );
		
		this.defaultCloseMode = defaultCloseMode;
		this.defaultCloseEnableMode = defaultCloseEnableMode;
		this.defaultCoaOnEnable = defaultCoaOnEnable;
		this.defaultDisableServicesOnClose = defaultDisableServicesOnClose;
		this.defaultCloseRemoveFromKeyMap = defaultCloseRemoveFromKeyMap;
	}

	@Override
	public Object init( Setup setup, int moduleId, InetDevice device, InetDeviceType deviceType, ParameterMap deviceConfig )
	        throws Exception
	{
		super.init( setup, moduleId, device, deviceType, deviceConfig );
		
		this.closeMode = deviceConfig.getInt( "sa.radius.connection.close.mode", defaultCloseMode );
		this.closeEnableMode = deviceConfig.getInt( "sa.radius.connection.close.enableMode", defaultCloseEnableMode );
		this.closeDisableMode = deviceConfig.getInt( "sa.radius.connection.close.disableMode", closeMode );
		
		// нужно ли слать CoA на включение сервисов при переключении из ограничнного доступа
		// по умолчанию просто ждем следующего DHCP-REQUEST чтобы ответить NAK и выдать былый адрес
		this.coaOnEnable = deviceConfig.getInt( "sa.radius.connection.coa.onEnable", defaultCoaOnEnable ? 1 : 0 ) > 0;

		// при отключечнии соединения по умолчанию убирать из keymap (т.е. начинать посылать NAK на DHCP запросы)
		this.closeRemoveFromKeyMap = deviceConfig.getInt( "sa.radius.connection.close.removeFromKeyMap", defaultCloseRemoveFromKeyMap ? 1 : 0 ) > 0;

		// при отключении соединения по умолчанию отключать сервисы
		this.disableServicesOnClose = deviceConfig.getInt( "sa.radius.connection.close.disableServices", defaultDisableServicesOnClose ? 1 : 0 ) > 0;

		// набор атрибутов для закрытия всех сервисов sa.radius.service.closeAttributes=
		this.serviceCloseAttributes = RadiusAttributeSet.newRadiusAttributeSet( deviceConfig.get( "sa.radius.service.closeAttributes", deviceConfig.get( "close.attributes", "Deactivate-Service-Name:1=RSE-SVC-EXT" ) ) );

		return null;
	}

	@Override
	public Object destroy()
	        throws Exception
	{
		super.destroy();

		return null;
	}

	/**
	 * Изменение состояния сессии или набора опций.
	 */
	@Override
	public Object connectionModify( ServiceActivatorEvent e )
	        throws Exception
	{
		logger.info( "Connection modify: oldState: " + e.getOldState() + "; newState: " + e.getNewState() + "; oldOptionSet: " + e.getOldOptions() + "; newOptionSet: " + e.getNewOptions() );
		
		final InetConnection connection = e.getConnection();
		
		// если нужно переключить в состояние отключен
		if( e.getNewState() == InetServ.STATE_DISABLE )
		{
			// если всегда сбрасываем соединение при переключении состояния - вызываем просто connectionClose
			// например, чтобы поменять IP с белого на серый
			// sa.radius.connection.withoutBreak=0
			if( !withoutBreak )
			{
				return connectionClose( connection, closeDisableMode, disableServicesOnClose );
			}
			
			// устанавливаем флаг, что нужно будет поменять состояние соединения в базе
			// sa.radius.connection.stateModify=1
			if( needConnectionStateModify )
			{
				e.setConnectionStateModified( true );
			}

			// отключаем соединение
			return connectionDisable( connection );
		}

		// если раньше было состояние отключено, а теперь включено
		if( e.getOldState() == InetServ.STATE_DISABLE )
		{
			// если всегда сбрасываем соединение при переключении состояния - вызываем просто connectionClose
			// например, чтобы поменять IP с белого на серый
			// sa.radius.connection.withoutBreak=0
			if( !withoutBreak )
			{
				Object result = connectionClose( connection, closeEnableMode, false );

				// если при необходимости включить доступ до того как поменяли IP мы хотим включить доступ на SE, то идем дальше
				// иначе return
				if( !coaOnEnable )
				{
					return result;
				}
			}
			
			// устанавливаем флаг, что нужно будет поменять состояние соединения в базе
			// sa.radius.connection.stateModify=1
			if( needConnectionStateModify )
			{
				e.setConnectionStateModified( true );
			}
		}

		// закрываем все сервисы
		if( !serviceCloseAttributes.isEmpty() )
		{
			RadiusPacket request = radiusClient.createModifyRequest();
			prepareRequest( request, connection );
			request.addAttributes( serviceCloseAttributes );

			logger.info( "Send close services CoA: \n" + request );

			radiusClient.send( request );
		}

		// готовим пакет на включение сервисов
		RadiusPacket request = radiusClient.createModifyRequest();
		prepareRequest( request, connection );
		
		// если раньше было состояние отключено, а теперь включено
		// добавляем атрибуты из sa.radius.enable.attributes=
		if( e.getOldState() == InetServ.STATE_DISABLE )
		{
			request.addAttributes( enableRadiusAttributes );
		}
		
		final String realm = e.getRealm();

		// открываем все подключенные сервисы
		// используя атрибуты опций из конфига устройства с префиксом sa.radius.option.attributesPrefix=radius.inetOption.
		// (т.е. по умолчанию атрибуты radius.inetOption.<id_опции>.attributes=)
		// и/или конфига опций
		for( Integer option : e.getNewOptions() )
		{
			RadiusAttributeSet set = optionRadiusAttributesMap.get( realm, option );
			if( set != null )
			{
				request.addAttributes( set );
			}
		}

		logger.info( "Send enable services CoA: \n" + request );

		return radiusClient.sendAsync( request );
	}
	
	/**
	 * Ограничение доступа.
	 */
	protected Object connectionDisable( InetConnection connection )
	        throws Exception
	{
		logger.info( "Connection disable" );

		RadiusPacket request = radiusClient.createModifyRequest();
		prepareRequest( request, connection );

		// выключение сервиса, отправляем атрибуты редиректа
		// sa.radius.disable.attributes= radius.disable.attributes=
		request.addAttributes( disableRadiusAttributes );

		logger.info( "Send CoA lock: \n" + request );

		return radiusClient.sendAsync( request );
	}
	
	/**
	 * Закрытие соединения по указанному режиму.
	 * @see #CLOSE_MODE_NONE
	 * @see #CLOSE_MODE_POD
	 * @param connection
	 * @param closeMode
	 * @param result
	 * @return
	 * @throws Exception
	 */
	protected Object connectionClose( final InetConnection connection, final int closeMode,
									  final boolean disableServices )
		throws Exception
	{
		logger.info( "Connection close mode " + closeMode );
		
		Object result;

		if( closeRemoveFromKeyMap )
		{
			// убрать из DHCP, чтобы выдало NaK
			logger.debug( "Remove connection from key map." );
			EventProcessor.getInstance().request( new InetConnectionManager.ConnectionRemoveEvent( connection ) );
		}

		if( disableServices )
		{
			result = connectionDisable( connection );
		}
		else
		{
			result = null;
		}

		switch( closeMode )
		{
			default:
			case CLOSE_MODE_NONE:
			{
				break;
			}

			// отправляем PoD-пакет
			case CLOSE_MODE_POD:
			{
				RadiusPacket request = radiusClient.createDisconnectRequest();
				prepareRequest( request, connection );

				logger.info( "Send PoD: \n" + request );
				result = radiusClient.sendAsync( request );

				break;
			}
		}

		return result;
	}

	/**
	 * Сброс соединения.<br/>
	 * {@inheritDoc}
	 */
	@Override
	public Object connectionClose( final ServiceActivatorEvent e )
		throws Exception
	{
		logger.info( "Connection close" );

		final InetConnection connection = e.getConnection();

		return connectionClose( connection, closeMode, disableServicesOnClose );
	}
}
